from .base import BaseHttp


class BlockingHttp(BaseHttp):
    def _get(self, url, read_bytes: bool = False,
             read_json: bool = True, *args, **kwargs) -> dict:
        """Wrapped HTTPX Delete.
        """

        return self.handle_resp(
            self._client.get(url, *args, **kwargs),
            json=read_json,
            read=read_bytes
        )

    def _delete(self, url, *args, **kwargs) -> bool:
        """Wrapped HTTPX Delete.
        """

        return self.handle_resp(
            self._client.delete(url, *args, **kwargs),
            False
        )

    def _post(self, url, read_json: bool = False,
              *args, **kwargs) -> bool:
        """Wrapped HTTPX Post.
        """

        return self.handle_resp(
            self._client.post(url, *args, **kwargs),
            read_json
        )

    def _put(self, url, *args, **kwargs) -> bool:
        """Wrapped HTTPX Put.
        """

        return self.handle_resp(
            self._client.put(url, *args, **kwargs),
            False
        )



import typing


class MapsModel:
    """Holds map details

    Attributes
    ----------
    map : str
    seconds : int
    """

    def __init__(self, data: dict) -> None:
        self.map = data["map"]
        self.seconds = data["seconds"]


class PlayerModel:
    """Holds player details

    Attributes
    ----------
    name : str
    duration : int
    score : int
    """

    def __init__(self, data: dict) -> None:
        self.name = data["name"]
        self.duration = data["duration"]
        self.score = data["score"]


class PlayersOnlineGraphModel:
    """Holds player graph details

    Attributes
    ----------
    timestamp : str
    value : str
    """

    def __init__(self, data: dict) -> None:
        self.timestamp = data["timestamp"]
        self.value = data["value"]


class MetricsModel:
    def __init__(self, data: dict) -> None:
        self.__data = data

    def maps(self) -> typing.Generator[MapsModel, None, None]:
        """Used to list all maps what have been played.

        Yields
        -------
        MapsModel
            Holds details on maps.
        """

        for data in self.__data["maps_played"]:
            yield MapsModel(data)

    def players_online(self) -> typing.Generator[PlayerModel, None, None]:
        """Used to list all players online.

        Yields
        -------
        PlayerModel
            Holds details on online players.
        """

        for data in self.__data["players_online"]:
            yield PlayerModel(data)

    def players_online_graph(self) -> \
            typing.Generator[PlayersOnlineGraphModel, None, None]:
        """Used to list all players online graph.

        Yields
        -------
        PlayersOnlineGraphModel
            Holds details on online player times.
        """

        for data in self.__data["players_online_graph"]:
            yield PlayersOnlineGraphModel(data)

    def all_time_players(self) -> typing.Generator[PlayerModel, None, None]:
        """Used to list all time players.

        Yields
        -------
        PlayerModel
            Holds details on online players.
        """

        for data in self.__data["all_time_players"]:
            yield PlayerModel(data)



import unittest

from secrets import token_urlsafe

from ..models.account import AccountModel
from ..models.server import ServerModel
from ..models.file import FileModel
from ..models.backup import BackupModel
from ..models.metrics import MetricsModel, MapsModel, PlayerModel, \
    PlayersOnlineGraphModel
from ..models.match import MatchModel, TeamModel, MatchPlayerModel

from ..server.blocking.backup import BlockingBackup
from ..server.blocking.file import BlockingFile

from ..match.blocking import BlockingMatch

from ..server.blocking import ServerBlocking

from ..settings import ServerSettings, MatchSettings

from .. import Blocking

from .shared_vars import EMAIL, PASSWORD, TEST_IMAGE_DIRETORY


class TestBlockingClient(unittest.TestCase):
    def setUp(self):
        self.client = Blocking(
            email=EMAIL,
            password=PASSWORD,
            timeout=360
        )

    def tearDown(self):
        self.client.close()

    def test_blocking_account(self):
        account = self.client.account()

        self.assertTrue(isinstance(account, AccountModel))

    def test_blocking_domains(self):
        for domain in self.client.domains():
            self.assertTrue(type(domain) == str)

    def test_blocking_context(self):
        context = Blocking(EMAIL, PASSWORD)

        with context as client:
            client.account()

        with context as client:
            for domain in client.domains():
                pass

    def test_blocking_list_servers(self):
        for data, server in self.client.servers():
            self.assertIsInstance(data, ServerModel)
            self.assertIsInstance(server, ServerBlocking)

    def test_blocking_server_csgo(self):
        server_data, server = self.client.create_server(
            ServerSettings(
                name="Blocking CS: GO server",
                location="sydney",
            ).csgo(
                slots=5,
                game_token="",
                tickrate=128,
                rcon_password=token_urlsafe()
            )
        )

        self.assertIsInstance(server_data, ServerModel)
        self.assertIsInstance(server, ServerBlocking)

        self.assertIsInstance(server.get(), ServerModel)

        server.console_send("status")
        server.console_retrive()

        server.start()
        server.stop()
        server.reset()

        server.update(
            ServerSettings(
                name="Renamed Blocking CS: GO server",
                location="sydney"
            ).csgo(
                slots=7
            )
        )

        for data, f in server.files():
            self.assertIsInstance(data, FileModel)
            self.assertIsInstance(f, BlockingFile)

        for data, f in server.files(hide_default=True, file_sizes=True):
            self.assertIsInstance(data, FileModel)
            self.assertIsInstance(f, BlockingFile)

        for data, backup in server.backups():
            self.assertIsInstance(data, BackupModel)
            self.assertIsInstance(backup, BlockingBackup)

            backup.restore()

        metrics = server.metrics()
        self.assertIsInstance(metrics, MetricsModel)

        for map_ in metrics.maps():
            self.assertIsInstance(map_, MapsModel)

        for player in metrics.players_online():
            self.assertIsInstance(player, PlayerModel)

        for player in metrics.all_time_players():
            self.assertIsInstance(player, PlayerModel)

        for player in metrics.players_online_graph():
            self.assertIsInstance(player, PlayersOnlineGraphModel)

        server.ftp_reset()

        test_1_file = server.file("test.txt")
        self.assertIsNone(test_1_file.upload(b"hello world"))
        self.assertIsNone(test_1_file.move("cfg"))

        self.assertTrue(type(test_1_file.dowload()) == bytes)

        self.assertIsNone(test_1_file.delete())

        test_2_file = server.file("test.jpg")
        test_2_file.upload_file(TEST_IMAGE_DIRETORY)
        test_2_file.save(TEST_IMAGE_DIRETORY)

        self.assertIsNone(test_2_file.delete())

        _, duplicate = server.duplicate(sync=True)
        self.assertIsNone(duplicate.delete())

        match_data, match = server.create_match(
            MatchSettings(
            ).team_1(
                [
                    "[U:1:116962485]",
                    76561198017567105,
                    "STEAM_0:1:186064092"
                ]
            ).team_2(
                [
                    "[U:1:320762620]",
                    "STEAM_1:1:83437164",
                    76561198214871324
                ]
            )
        )

        self.assertIsInstance(match_data, MatchModel)
        self.assertIsInstance(match, BlockingMatch)

        self.assertIsInstance(match.get(), MatchModel)

        self.assertIsInstance(match_data.team_1, TeamModel)
        self.assertIsInstance(match_data.team_2, TeamModel)

        for player in match_data.players():
            self.assertIsInstance(player, MatchPlayerModel)

        self.assertIsNone(server.delete())

    def test_blocking_server_valheim(self):
        data, server = self.client.create_server(
            ServerSettings(
                name="Blocking valheim server",
                location="sydney"
            ).valheim(
                password=token_urlsafe(8),
                world_name="Poggers",
                plus=False,
                admins=[
                    "[U:1:116962485]",
                    76561198017567105,
                    "STEAM_0:1:186064092",
                    "76561198214871321"
                ]
            )
        )

        self.assertIsInstance(data, ServerModel)
        self.assertIsInstance(server, ServerBlocking)

        self.assertIsInstance(server.get(), ServerModel)

        self.assertIsNone(server.delete())

    def test_blocking_server_mumble(self):
        data, server = self.client.create_server(
            ServerSettings(
                name="Blocking Mumble server",
                location="sydney"
            ).mumble(
                slots=7,
                superuser_password=token_urlsafe()
            )
        )

        self.assertIsInstance(data, ServerModel)
        self.assertIsInstance(server, ServerBlocking)

        self.assertIsInstance(server.get(), ServerModel)

        self.assertIsNone(server.delete())

    def test_blocking_server_tf2(self):
        data, server = self.client.create_server(
            ServerSettings(
                name="Blocking TF2 server",
                location="sydney"
            ).tf2(
                slots=5,
                rcon_password=token_urlsafe()
            )
        )

        self.assertIsInstance(data, ServerModel)
        self.assertIsInstance(server, ServerBlocking)

        self.assertIsInstance(server.get(), ServerModel)

        self.assertIsNone(server.delete())

    def test_blocking_server_teamspeak(self):
        data, server = self.client.create_server(
            ServerSettings(
                name="Blocking TF2 server",
                location="sydney"
            ).teamspeak(
                slots=5,
            )
        )

        self.assertIsInstance(data, ServerModel)
        self.assertIsInstance(server, ServerBlocking)

        self.assertIsInstance(server.get(), ServerModel)

        self.assertIsNone(server.delete())



from datetime import datetime


class BackupModel:
    """Holds detail on backups.

    Attributes
    ----------
    backup_name : str
    timestamp : datetime.datetime
    """

    def __init__(self, data: dict) -> None:
        self.backup_name = data["name"]
        self.timestamp = datetime.strptime(
            data["timestamp"],
            "%a %b %d %H:%M %Y"
        )



import typing


class TeamModel:
    """Holds details on team.

    Attributes
    ----------
    score : int
    players : list
    """

    def __init__(self, players: list, score: int) -> None:
        self.score = score
        self.players = players


class MatchPlayerModel:
    """Holds match player details.

    Attributes
    ----------
    steamid : str
    kills : int
    deaths : int
    assists : int
    """
    def __init__(self, data: dict) -> None:
        self.steamid = data["steam_id"]
        self.kills = data["kills"]
        self.deaths = data["deaths"]
        self.assists = data["assists"]


class MatchModel:
    """Holds match details.

    Attributes
    ----------
    match_id : str
    server_id : str
    connect_time : int
    round_end_webhook : str
    match_end_webhook : str
    finished : bool
    cancel_reason : str
    rounds_played : int
    spectators : list
    team_1 : TeamModel
    team_2 : TeamModel
    knife_round : bool
    playwin : bool
    playwin_webhook : str
    playwin_result : dict
    warmup_time : int
    wait_for_spectators : bool
    """
    def __init__(self, data: dict) -> None:
        self.match_id = data["id"]
        self.server_id = data["game_server_id"]
        self.connect_time = data["connect_time"]
        self.round_end_webhook = data["round_end_webhook_url"]
        self.match_end_webhook = data["match_end_webhook_url"]
        self.finished = data["finished"]
        self.cancel_reason = data["cancel_reason"]
        self.rounds_played = data["rounds_played"]
        self.spectators = data["spectator_steam_ids"]
        self.knife_round = data["enable_knife_round"]
        self.playwin = data["enable_playwin"]
        self.playwin_webhook = data["playwin_result_webhook_url"]
        self.playwin_result = data["playwin_result"]
        self.warmup_time = data["warmup_time"]
        self.wait_for_spectators = data["wait_for_spectators"]

        self.team_1 = TeamModel(
            data["team1_steam_ids"],
            data["team1_stats"]["score"]
        )
        self.team_2 = TeamModel(
            data["team2_steam_ids"],
            data["team2_stats"]["score"]
        )

        self.__players = data["player_stats"]

    def players(self) -> typing.Generator[MatchPlayerModel, None, None]:
        """Used to list players.

        Yields
        -------
        PlayerModel
            Holds details on player.
        """

        for player in self.__players:
            yield MatchPlayerModel(player)



from .base import MatchBase

from ..routes import MATCHES

from ..models.match import MatchModel


class BlockingMatch(MatchBase):
    def get(self) -> MatchModel:
        """Gets details on a match

        Returns
        -------
        MatchModel
            Holds match details.
        """

        data = self.context._get(
            MATCHES.details.format(self.match_id)
        )

        return MatchModel(data)



class DathostException(Exception):
    """Base exception for dathost.
    """

    def __init__(self, msg="Dathost base exception", *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)


class InvalidSlotSize(DathostException):
    """Raised when slot size is invalid.
    """

    def __init__(self, msg="Invalid slot size", *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)


class MultipleGames(DathostException):
    """Raised when you attempt to create one server
       with multiple games.
    """

    def __init__(self, msg="Multiple games called", *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)


class InvalidTickrate(DathostException):
    """Raised when tickrate is invalid.
    """

    def __init__(self, msg="Invalid tickrate", *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)


class InvalidConsoleLine(DathostException):
    """Raised when console line is above 1 or above 100000.
    """

    def __init__(self, msg="Invalid console line", *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)


class AwaitingOnly(DathostException):
    """Raised when a coroutine called is awaiting supported only.
    """

    def __init__(self, msg="Only awaiting supported", *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)


class InvalidSteamID(DathostException):
    """Raised when give ID isn't understood.
    """

    def __init__(self, msg="Invalid Steam ID", *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)


class NotFound(DathostException):
    """Resource not found.
    """

    def __init__(self, msg="Resource not found", *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)


class BadRequest(DathostException):
    """Path is a directory or Cannot move file into itself.
    """

    def __init__(self, msg="Bad request", *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)


class ExceededStorage(DathostException):
    """Your disk quota of 30GB per server (excluding base installation)
    has been exceeded
    """

    def __init__(self, msg="Storage exceeded", *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)


class ServerStart(DathostException):
    """Failed to start server.
    """

    def __init__(self, msg="Failed to start server", *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)


class InvalidStorageSize(DathostException):
    """Storage size is invalid.
    """

    def __init__(self, msg="Storage size is invalid", *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)



class FileModel:
    """Used to hold details on file.

    Attributes
    ----------
    path : str
    size : str, optional
    """

    def __init__(self, data: dict) -> None:
        self.path = data["path"]
        self.size = data.get("size")



import typing

from .base import Base
from .routes import ACCOUNT, CUSTOM_DOMAINS, SERVER

from .http import AwaitingHttp, BlockingHttp

from .server.blocking import ServerBlocking
from .server.awaiting import ServerAwaiting

from .match.awaiting import AwaitingMatch
from .match.blocking import BlockingMatch

from .settings import ServerSettings

from .models.account import AccountModel
from .models.server import ServerModel

from httpx import AsyncClient, Client


__version__ = "0.1.11"
__url__ = "https://dathost.readthedocs.io/en/latest/"
__description__ = "Asynchronous / Synchronous dathost API wrapper."
__author__ = "WardPearce"
__author_email__ = "wardpearce@protonmail.com"
__license__ = "GPL v3"


class Awaiting(Base, AwaitingHttp):
    __client_closed = False

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.__create_client()

    async def __aenter__(self) -> None:
        if self.__client_closed:
            self.__create_client()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    def __create_client(self) -> None:
        self._client = AsyncClient(
            auth=self._basic_auth,
            timeout=self._timeout
        )

    async def close(self) -> None:
        """Closes sessions
        """

        self.__client_closed = True
        await self._client.aclose()

    def match(self, match_id: str) -> AwaitingMatch:
        """Used to interact with a match.

        Parameters
        ----------
        match_id : str
            Dathost Match ID.

        Returns
        -------
        AwaitingMatch
        """

        return AwaitingMatch(
            self,
            match_id
        )

    async def create_server(self, settings: ServerSettings
                            ) -> typing.Tuple[ServerModel, ServerAwaiting]:
        """Creates a new server.

        Parameters
        ----------
        settings : ServerSettings
            Used to configure server.

        Returns
        -------
        ServerModel
            Holds data on server.
        ServerAwaiting
            Used to interact with the created server.
        """

        data = await self._post(
            url=SERVER.create,
            read_json=True,
            data=settings.payload,
        )

        return ServerModel(data), self.server(data["id"])

    def server(self, server_id: str) -> ServerAwaiting:
        """Used for interacting with a server.

        Parameters
        ----------
        server_id : str
            Datahost server ID.

        Returns
        -------
        ServerAwaiting
            Used to interact with the server.
        """

        return ServerAwaiting(self, server_id)

    async def servers(self) -> typing.AsyncGenerator[
            ServerModel, ServerAwaiting]:
        """Used to list servers.

        Yields
        -------
        ServerModel
            Holds data on server.
        """

        for server in await self._get(SERVER.list):
            yield ServerModel(server), self.server(server["id"])

    async def account(self) -> AccountModel:
        """Gets account details

        Returns
        -------
        AccountModel
            Holds data on a account.
        """

        return AccountModel(
            await self._get(ACCOUNT.details)
        )

    async def domains(self) -> typing.AsyncGenerator[str, None]:
        """Used to list domains.

        Returns
        -------
        list
            List of domains.
        """

        data = await self._get(CUSTOM_DOMAINS.details)
        for domain in data:
            yield domain["name"]


class Blocking(Base, BlockingHttp):
    __client_closed = False

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.__create_client()

    def __enter__(self) -> None:
        if self.__client_closed:
            self.__create_client()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    def close(self) -> None:
        """Closes sessions
        """

        self.__client_closed = True
        self._client.close()

    def __create_client(self) -> None:
        self._client = Client(
            auth=self._basic_auth,
            timeout=self._timeout
        )

    def match(self, match_id: str) -> BlockingMatch:
        """Used to interact with a match.

        Parameters
        ----------
        match_id : str
            Dathost Match ID.

        Returns
        -------
        BlockingMatch
        """

        return BlockingMatch(
            self,
            match_id
        )

    def create_server(self, settings: ServerSettings
                      ) -> typing.Tuple[ServerModel, ServerBlocking]:
        """Creates a new server.

        Parameters
        ----------
        settings: ServerSettings
            Used to configure server.

        Returns
        -------
        ServerModel
            Holds data on server.
        ServerBlocking
            Used to interact with the created server.
        """

        data = self._post(
            url=SERVER.create,
            read_json=True,
            data=settings.payload,
        )

        return ServerModel(data), self.server(data["id"])

    def server(self, server_id: str) -> ServerBlocking:
        """Used for interacting with a server.

        Parameters
        ----------
        server_id : str
            Datahost server ID.

        Returns
        -------
        ServerBlocking
            Used to interact with the server.
        """

        return ServerBlocking(self, server_id)

    def servers(self) -> typing.Generator[ServerModel, ServerBlocking, None]:
        """Used to list servers.

        Yields
        -------
        ServerModel
            Holds data on server.
        """

        for server in self._get(SERVER.list):
            yield ServerModel(server), self.server(server["id"])

    def account(self) -> AccountModel:
        """Gets account details

        Returns
        -------
        AccountModel
            Holds data on a account.
        """

        return AccountModel(
            self._get(ACCOUNT.details)
        )

    def domains(self) -> typing.Generator[str, None, None]:
        """Used to list domains.

        Returns
        -------
        list
            List of domains.
        """

        data = self._get(CUSTOM_DOMAINS.details)
        for domain in data:
            yield domain["name"]



from httpx import BasicAuth


class Base:
    def __init__(self, email: str, password: str, timeout: int = 60) -> None:
        """Used to create Dathost basic auth.

        Parameters
        ----------
        email : str
            Email of dathost account.
        password : str
            Password of dathost account.
        timeout : int, optional
            by default 60
        """

        self._basic_auth = BasicAuth(email, password)
        self._timeout = timeout
